""" Database connection """

__version__ = "$Revision: 1.23 $"
# $Source: /home/vladap/src/mysqlUserFolder/RCS/db.py,v $
#
# $Author: vladap $
# $Date: 2003/01/20 21:03:23 $

import string

import util, cfg

# --------------------------------------------------------------------- */
# Utility functions


def setup_user_roles (roles):
	roles.append (cfg.PGSQL_USER_ROLE)
	roles.append (cfg.ZOPE_ANONYMOUS_ROLE)
	roles.append (cfg.ZOPE_AUTHENTICATED_ROLE)
			
	iroles = cfg.VALIDATE_IGNORE_ROLES
	
	if len (iroles) > 0:
		new_roles = filter (lambda x, iroles=iroles: not (x in iroles), roles)
	else:
		new_roles = roles

	return new_roles

# --------------------------------------------------------------------- */
# dbConnection class

class dbConnection:
    """ DB Connection class. 
    
        It is supposed to be refferenced as _v_ attribute of all classes
        that need access to a database. So this class does not use locks.
    """  

    def __init__ (Self, DSN, dbtype):
        Self.DSN = DSN
        Self.dbtype = dbtype
        Self.db_con = None
    #--------------------------------------------
    # Internal general utility functions
    def __query_list_field (Self, cur, query, args):
        """ This function returns a list of values from the select query.
            Result set should contain only one field. """
        Self.execute (cur, query, args)
        data = cur.fetchall ()
        ret = []
        
        for i in range (len (data)):
            (value,) = data [i]
            ret.append (value)
        
        return ret

    def __get_nextval(Self, cur, seq_name):
        Self.execute(cur, "SELECT nextval(%s) ", (seq_name, ))
        return cur.fetchone()[0]
        
    def __query_count (Self, cur, table, where_clause = None, args = None):
        """ Returns number of rows in the recordset specified
            by where clause. 
        """
        query = 'select count(*) from %s' % (table)
        if where_clause:
            query = '%s where %s' % (query, where_clause)
            
        Self.execute (cur, query, args)
        
        data = cur.fetchall ()
        
        return data [0][0]
        
    #--------------------------------------------
    # Connect / dicconnect 
    
    def connect (Self):
        import psycopg
        util.log_debug ("db.connect (): " + Self.DSN)
        Self.db_con = psycopg.connect (dsn = Self.DSN, serialize=0)
        return 1
    
    def disconnect (Self):
        if not Self.db_con is None:
            Self.db_con.close ()
            
        Self.db_con = None
    
    #--------------------------------------------
    # Cursor / execute / commit
    def cursor (Self):
        if Self.db_con is None:
            Self.connect()
        return Self.db_con.cursor ()

    def execute(Self, cur, query, param):
        util.log_debug ("execute(): " + query + str(param))
        return cur.execute (query, param)

    def commit (Self, cur):
        cur.commit ()
    
    def rollback (Self, cur):
        cur.rollback ()
    
    #--------------------------------------------
    # Internal User db functions.
    
    def __check_user (Self, user):
        username = user.getUserName ()
        realm = user.getRealm ()
        id = user.getDBId ()
        
        where = "Username = %s and Realm = %s and Id = %s"
                                                
        cur = Self.cursor ()
        cnt = Self.__query_count (cur, "ZUsers", where, (username, realm, id))
        Self.commit (cur)
        
        return (cnt == 1)
        
    def __check_user_exist (Self, username, realm):
        where = "Username = %s and Realm = %s"
        cur = Self.cursor ()
        cnt = Self.__query_count (cur, "ZUsers", where, (username, realm))
        Self.commit (cur)
        
        return (cnt == 1)
        
    #--------------------------------------------
    # User db API functions.
        
    def list_rolenames (Self, realm):
        util.log_debug ("list_rolenames()")
        cur = Self.cursor ()
        query = "select Rolename from ZRoles where Realm = %s order by Rolename"
        ret = Self.__query_list_field (cur, query, (realm, ))
        Self.commit (cur)
        
        return ret
        
    def list_usernames (Self, realm):
        util.log_debug ("list_usernames()")
        cur = Self.cursor ()
        query = "select Username from ZUsers where Realm = %s order by Username"
        ret = Self.__query_list_field (cur, query, (realm, ))
        Self.commit (cur)
        
        return ret


    def list_usernames_with_role (Self, realm, role):
        util.log_debug ("list_rolenames_with_role()")
        cur = Self.cursor ()
        query = "Select id from ZRoles where Realm = %s and Rolename = %s"
        [rid, ] = Self.__query_list_field (cur, query, (realm, role))

        query = "Select ZUsers.Username from ZUsers,ZUserRoles where " \
                    "(ZUsers.Id = ZUserRoles.Id_User) and (ZUserRoles.Id_Role = %s)" \
                    "and (ZUsers.Realm = %s) order by Username"
        ret = Self.__query_list_field (cur, query, (rid, realm))
        Self.commit (cur)
        return ret
        
    def get_user_data (Self, user, by_id = 0):
        util.log_debug ("get_user_data()")
        username = user.getUserName ()
        realm = user.getRealm ()
        
        if by_id: uid = user.getDBId ()
        
        user._setDBId (cfg.VDB_INVALID_ID)
        user._setPassword ('', 0)
        user._setAuthData ([], [])
        user._setOtherData ('<unknown>', '<unknown>')
    
            # ----- Trazenje usera
        cur = Self.cursor ()
            
        if by_id:
            query = "Select Username, Realm from ZUsers where Id = %s"
            Self.execute (cur, query, (uid, ))
            data = cur.fetchall ()
            if len (data) == 0: 
                util.log_debug ("dbConn.get_user_data (): " 
                                            "no user data by id: '%i'" % uid)
                Self.commit (cur)
                return 0
                
            (du, dr) = data [0]
            username = du
            user._setUserName (username)
            
            if dr != realm:
                util.log_debug ("dbConn.get_user_data (): "
                                                "invalid realm for user id")
                return 0
            
        query = "Select Id, Password, PasswordType, EMail, Realname, " \
                    "Domains from ZUsers where Realm = %s and Username = %s"
        
        Self.execute (cur, query, (realm, username))
        data = cur.fetchall ()
        if len (data) == 0: 
            util.log_debug ("dbConn.get_user_data (): no user data")
            Self.commit (cur)
            return 0
                  
        (id, password, password_type, email, realname, domains_str) = data [0]

        domains = string.split (string.strip (domains_str))
        
            # ----- Find roles 
        
        query = 'Select Id_Role from ZUserRoles where Id_User = %i'
        role_ids = Self.__query_list_field (cur, query, (id, ))
    
        roles = []
        
        for role_id in role_ids:
            query = "Select Rolename from ZRoles where Id = %i"
            (role,) = Self.__query_list_field (cur, query, (role_id, ))
            
            roles.append (role)
            
        roles = setup_user_roles (roles)
        
            # ----- Popunjavanje strukture

            
        user._setDBId (id)
        user._setOtherData (realname, email)
        user._setAuthData (roles, domains)
        user._setPassword (password, password_type)
        
        Self.commit (cur)
        return 1
        
    def count_users (Self, realm):
        util.log_debug ("count_users()")
        where = "Realm = %s"
        table = "ZUsers"
        cur = Self.cursor ()
        cnt = Self.__query_count (cur, table, where, (realm, ))
        Self.commit (cur)
        return cnt
        
    def count_roles (Self, realm):
        util.log_debug ("count_roles()")
        where = "Realm = %s"
        table = "ZRoles"
        cur = Self.cursor ()
        cnt = Self.__query_count (cur, table, where, (realm, ))
        Self.commit (cur)
        return cnt
        
    def change_user_password (Self, user, password, ptype):
        util.log_debug ("change_user_password()")
        if not Self.__check_user (user):
            return 0
    
        id = user.getDBId ()
    
        cur = Self.cursor ()
        query = "update ZUsers set " \
            "Password = %s, PasswordType = %s where Id = %s"
                                
        Self.execute (cur, query, (password, ptype, id))
        ret = (cur.rowcount == 1)
            
        Self.commit(cur)
        return ret

    def change_user_roles (Self, user, roles):
        util.log_debug ("change_user_roles()")
        if not Self.__check_user (user):
            return 0

        user_id = user.getDBId ()
        realm = user.getRealm ()
        
        cur = Self.cursor ()
        query = "delete from ZUserRoles where Id_User = %i"
        Self.execute (cur, query, (user_id,))

        for role in roles:
            query = \
                "Select Id from ZRoles where Rolename = %s and Realm = %s"

            ret = Self.__query_list_field (cur, query, (role, realm))
            if ret != []:
                [role_id] = ret
            
                query = \
                    "insert into ZUserRoles (Id_Role, Id_User) Values (%s, %s)"
                Self.execute (cur, query, (role_id, user_id))
            else:
                Self.rollback (cur)
                raise 'Unknown role: %s' % (query)

        Self.commit (cur)
        return 1

    def change_user_domains (Self, user, domains):
        util.log_debug ("change_user_domains()")
        if type (domains) == type ([]):
            domains = string.join (domains)
            
        if not Self.__check_user (user):
            return 0
    
        id = user.getDBId ()

        cur = Self.cursor()
        query = "update ZUsers set Domains = %s where Id = %s"
        Self.execute (cur, query, (domains, id))
        ret = (cur.rowcount == 1)
        
        Self.commit(cur)
        return ret
        
    def change_user_other (Self, user, realname, email):
        util.log_debug ("change_user_other()")
        if not Self.__check_user (user):
            return 0
    
        id = user.getDBId ()
    
        cur = Self.cursor()
        query = "update ZUsers set " \
            "Realname = %s, EMail = %s where Id = %s"
        Self.execute (cur, query, (realname, email, id))
        ret = (cur.rowcount == 1)
        
        Self.commit(cur)
        return ret
        
    def create_user (Self, username, realm, roles):
        util.log_debug ("create_user()")
        if Self.__check_user_exist (username, realm):
            return 0

        cur = Self.cursor()
        nextval = Self.__get_nextval(cur, 'ZUsers_seq')
        query = "insert into ZUsers (Id, Username, Realm, Password, " \
                        "PasswordType) values (%i, %s, %s, %s, %s)"

        Self.execute (cur, query, 
                (nextval, username, realm, '', cfg.VDB_INVALID_PASSWORD_TYPE))
        user_id = nextval

        for role in roles:
            query = \
                "Select Id from ZRoles where Rolename = %s and Realm = %s"

            ret = Self.__query_list_field (cur, query, (role, realm))
            if ret != []:
                [role_id] = ret
            
                query = \
                    "insert into ZUserRoles (Id_Role, Id_User) Values (%s, %s)"
                Self.execute (cur, query, (role_id, user_id))
            else:
                Self.del_user (username, realm)
                Self.commit (cur)
                raise 'Unknown role: (%s, %s)' % (role, realm)
        
        Self.commit(cur)
        return 1
        
    def del_user (Self, username, realm, mclass = None):
        util.log_debug ("del_user()")
        cur = Self.cursor ()
        query = "select Id from ZUsers where Username = %s and Realm = %s"
                    
        ret = Self.__query_list_field (cur, query, (username, realm))
        if ret == []:
            return 0
            
        [id] = ret

        query = "select Id from ZSessions where Id_User = %s"
        sids = Self.__query_list_field (cur, query, (id, ))
        
        for sid in sids:
            Self.del_token ('session', sid, realm)
        
        Self.del_token ('user', id, realm)
        
        query = "delete from ZSessions where Id_User = %s"
        Self.execute (cur, query, (id, ))
        
        query = "delete from ZUserRoles where Id_User = %s"
        Self.execute (cur, query, (id, ))
        
        query = "delete from ZUsers where Id = %s"
        Self.execute (cur, query, (id, ))
        Self.commit(cur)
        
        if mclass: Self.del_misc_data_class (mclass, id)
        
        return 1

    def create_role (Self, role, realm):
        util.log_debug ("create_role()")
        cur = Self.cursor ()
        nextval = Self.__get_nextval(cur, 'ZRoles_seq')
        query = "insert into ZRoles (Id, Rolename, Realm) values (%i, %s, %s)"
        Self.execute (cur, query, (nextval, role, realm))
        Self.commit(cur)
        return 1

    def del_role (Self, role, realm):
        util.log_debug ("del_role()")
        cur = Self.cursor ()
        query = "Select Id from ZRoles where Rolename = %s and Realm = %s"
        ret = Self.__query_list_field (cur, query, (role, realm))

        if ret == []:
            Self.commit (cur)
            return 0

        [role_id] = ret
            
        query = "delete from ZRoles where Id = %s"
        Self.execute (cur, query, (role_id,))

        query = "delete from ZUserRoles where Id_Role = %s"
        Self.execute (cur, query, (role_id,))
        Self.commit(cur)
        
        return 1

        
    #--------------------------------------------
    # Session db API functions
        
    def create_session (Self, session):
        util.log_debug ("create_session()")
        user_id = session.getUserId ()
        
        cur = Self.cursor ()
        nextval = Self.__get_nextval(cur, 'ZSessions_seq')
        query = "insert into ZSessions (Id, id_user) values (%i, %s)"
        Self.execute (cur, query, (nextval, user_id, ))
        Self.commit(cur)
        
        session._setId (nextval)
        
    def get_session (Self, sess):
        util.log_debug ("get_session()")
        id = sess.getId ()
        
        cur = Self.cursor ()
        query = "select Id_User from ZSessions where Id = %s"
        res = Self.__query_list_field (cur, query, (id, ))
        Self.commit (cur)
        
        if not res:
            sess._setId (cfg.VDB_INVALID_ID)
            sess._setUserId (cfg.VDB_INVALID_ID)
            
            return sess
            
        assert (len (res) == 1)
        sess._setUserId (res [0])
        
    def session_set_user_id (Self, session, user_id):
        util.log_debug ("session_set_user_id()")
        cur = Self.cursor ()
        query = "update ZSessions set Id_User = %s where Id = %s"
        
        Self.execute (cur, query, (user_id, session.getId ()))
        
        Self.commit(cur)
    
    #--------------------------------------------
    # Token db API functions
    
    
    def del_token (Self, class_id, token_id, realm):
        util.log_debug ("del_token()")
        cur = Self.cursor ()
        query = "delete from ZTokens where Class = %s " \
                                            "and Id = %s and Realm = %s"
        
        Self.execute (cur, query, (class_id, token_id, realm))

        Self.commit(cur)
        
    def create_token (Self, class_id, token_id, token_value, 
            token_timeout, token_life, created_time, accessed_time, realm):
        util.log_debug ("create_token()")
        Self.del_token (class_id, token_id, realm)
        
        cur = Self.cursor ()
        query = "insert into ZTokens (Class, Id, Value, Created, " \
                        "Accessed, Timeout, Life, Realm) values " \
                                        "(%s, %s, %s, %s, %s, %s, %s, %s)"
                                
        values = (class_id, token_id, token_value, created_time,
                            accessed_time, token_timeout, token_life, realm)
        
        Self.execute (cur, query,values)
        
        Self.commit (cur)
        
    def get_token (Self, class_id, token_id, realm):
        util.log_debug ("get_token()")
        cur = Self.cursor ()
        query = "select Value, Created, Accessed, Timeout, Life " \
                                    "from ZTokens where Class = %s " \
                                    "and Id = %s and Realm = %s"
                        
        Self.execute (cur, query, (class_id, token_id, realm))
        data = cur.fetchall ()
        Self.commit (cur)
        
        if len (data) == 0: return None
        assert (len (data) == 1)
        
        return data [0]
        
    def update_token_accessed (Self, class_id, token_id, time, realm):
        util.log_debug ("update_token_accessed()")
        cur = Self.cursor ()
        query = "update ZTokens set Accessed = %s where Class = %s " \
                                            "and Id = %s and Realm = %s"
        
        Self.execute (cur, query, (time, class_id, token_id, realm))

        Self.commit(cur)
        
    #--------------------------------------------
    # MiscData API functions
    
    def save_misc_data (Self, obj_class, obj_id, misc_id, misc_tuple):
        util.log_debug ("save_misc_data()")
        if type (misc_tuple) != type (()):
            raise 'API Error', \
                "save_misc_data () should be called with tuple. "
    
        (iv, sv) = misc_tuple
        
        cur = Self.cursor ()
        data = Self.get_misc_data (obj_class, obj_id, misc_id, check=1)
        if data:
            query = "update ZMiscData set " \
                    "Value_int = %s, Value_string = %s where " \
                    "ObjId = %s and OClass = %s and MId = %s" 
            
            Self.execute (cur, query, (iv, sv, obj_id, obj_class, misc_id))
            
        else:
            query = "insert into ZMiscData (ObjId, OClass, MId, " \
                "Value_int, Value_string) values (%s, %s, %s, %s, %s)"
                        
            Self.execute (cur, query, (obj_id, obj_class, misc_id, iv, sv))

        Self.commit(cur)
        
    def get_misc_data (Self, oclass, oid, mid, check = 0):
        util.log_debug ("get_misc_data()")
        cur = Self.cursor ()
        query = "select Value_int, Value_string from ZMiscData where " \
                                    "ObjId = %s and OClass = %s and MId = %s"
        
        Self.execute (cur, query, (oid, oclass, mid))
        
        data = cur.fetchall ()
        Self.commit (cur)

        if not data: 
            if not check: return cfg.MISC_DATA_EMPTY
            else: return 0
            
        assert (len (data) == 1)
        r = data [0]
        
        if check: return 1
        
        return r
        
    def del_misc_data (Self, oclass, oid, mid):
        util.log_debug ("del_misc_data()")
        cur = Self.cursor ()
        query = "delete from ZMiscData where " \
                                    "ObjId = %s and OClass = %s and MId = %s"
        
        Self.execute (cur, query, (oid, oclass, mid))

        Self.commit(cur)
        
    def del_misc_data_class (Self, oclass, oid):
        util.log_debug ("del_misc_data_class()")
        cur = Self.cursor ()
        query = "delete from ZMiscData where ObjId = %s and OClass = %s"
        
        Self.execute (cur, query, (oid, oclass))

        Self.commit(cur)
        
    #--------------------------------------------
    # Session logging
    
    def log_sql_simple (Self, realm, sid, event):
        util.log_debug ("log_sql_simple()")
        cur = Self.cursor ()
        nextval = Self.__get_nextval(cur, "ZLog_seq")
        query = "insert into ZLog (Id, Realm, Id_Session, Event) " \
                                               "values (%i, %s, %s, %s)"
        Self.execute (cur, query, (nextval, realm, sid, event))

        Self.commit(cur)
        
    def log_sql (Self, realm, sid, event, e_value, e_value2, o_type, 
                                                            o_size, r_addr):
        util.log_debug ("log_sql()")
        cur = Self.cursor ()
        nextval = Self.__get_nextval(cur, "ZLog_seq")
        query = "insert into ZLog (Id, Realm, Id_Session, Event, " \
            "EValue, EValue2, OType, OSize, RAddr) values " \
            "(%i, %s, %s, %s, %s, %s, %s, %s, %s)"
        
        Self.execute (cur, query, (nextval, realm, sid, event, 
                                    e_value, e_value2, o_type, o_size, r_addr))
        
        Self.commit (cur)
        
